@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2016-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2016-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -22,24 +22,34 @@
 import os.path
 import html
 import collections
+import functools
+import typing
 
-import sip
+import attr
 from PyQt5.QtCore import (pyqtSlot, pyqtSignal, Qt, QTimer, QDir, QModelIndex,
                           QItemSelectionModel, QObject, QEventLoop)
 from PyQt5.QtWidgets import (QWidget, QGridLayout, QVBoxLayout, QLineEdit,
-                             QLabel, QFileSystemModel, QTreeView, QSizePolicy)
+                             QLabel, QFileSystemModel, QTreeView, QSizePolicy,
+                             QSpacerItem)
 
 from qutebrowser.browser import downloads
-from qutebrowser.config import style, config
+from qutebrowser.config import config, configtypes, configexc, stylesheet
 from qutebrowser.utils import usertypes, log, utils, qtutils, objreg, message
 from qutebrowser.keyinput import modeman
-from qutebrowser.commands import cmdutils, cmdexc
+from qutebrowser.api import cmdutils
+from qutebrowser.utils import urlmatch
 
 
-prompt_queue = None
+prompt_queue = typing.cast('PromptQueue', None)
 
 
-AuthTuple = collections.namedtuple('AuthTuple', ['user', 'password'])
+@attr.s
+class AuthInfo:
+
+    """Authentication info returned by a prompt."""
+
+    user = attr.ib()
+    password = attr.ib()
 
 
 class Error(Exception):
@@ -47,7 +57,7 @@ class Error(Exception):
     """Base class for errors in this module."""
 
 
-class UnsupportedOperationError(Exception):
+class UnsupportedOperationError(Error):
 
     """Raised when the prompt class doesn't support the requested operation."""
 
@@ -92,8 +102,9 @@ class PromptQueue(QObject):
         super().__init__(parent)
         self._question = None
         self._shutting_down = False
-        self._loops = []
-        self._queue = collections.deque()
+        self._loops = []  # type: typing.MutableSequence[qtutils.EventLoop]
+        self._queue = collections.deque(
+        )  # type: typing.Deque[usertypes.Question]
         message.global_bridge.mode_left.connect(self._on_mode_left)
 
     def __repr__(self):
@@ -109,10 +120,11 @@ class PromptQueue(QObject):
         log.prompt.debug("Popping from queue {}".format(self._queue))
         if self._queue:
             question = self._queue.popleft()
-            if not sip.isdeleted(question):
-                # the question could already be deleted, e.g. by a cancelled
+            if not question.is_aborted:
+                # the question could already be aborted, e.g. by a cancelled
                 # download. See
-                # https://github.com/qutebrowser/qutebrowser/issues/415
+                # https://github.com/qutebrowser/qutebrowser/issues/415 and
+                # https://github.com/qutebrowser/qutebrowser/issues/1249
                 self.ask_question(question, blocking=False)
 
     def shutdown(self):
@@ -163,7 +175,7 @@ class PromptQueue(QObject):
             # just queue it up for later.
             log.prompt.debug("Adding {} to queue.".format(question))
             self._queue.append(question)
-            return
+            return None
 
         if blocking:
             # If we're blocking we save the old question on the stack, so we
@@ -184,7 +196,9 @@ class PromptQueue(QObject):
             question.completed.connect(loop.quit)
             question.completed.connect(loop.deleteLater)
             log.prompt.debug("Starting loop.exec_() for {}".format(question))
-            loop.exec_(QEventLoop.ExcludeSocketNotifiers)
+            flags = typing.cast(QEventLoop.ProcessEventsFlags,
+                                QEventLoop.ExcludeSocketNotifiers)
+            loop.exec_(flags)
             log.prompt.debug("Ending loop.exec_() for {}".format(question))
 
             log.prompt.debug("Restoring old question {}".format(old_question))
@@ -199,6 +213,7 @@ class PromptQueue(QObject):
             return question.answer
         else:
             question.completed.connect(self._pop_later)
+            return None
 
     @pyqtSlot(usertypes.KeyMode)
     def _on_mode_left(self, mode):
@@ -233,29 +248,38 @@ class PromptContainer(QWidget):
     """
 
     STYLESHEET = """
-        {% set prompt_radius = config.get('ui', 'prompt-radius') %}
         QWidget#PromptContainer {
-            {% if config.get('ui', 'status-position') == 'top' %}
-                border-bottom-left-radius: {{ prompt_radius }}px;
-                border-bottom-right-radius: {{ prompt_radius }}px;
+            {% if conf.statusbar.position == 'top' %}
+                border-bottom-left-radius: {{ conf.prompt.radius }}px;
+                border-bottom-right-radius: {{ conf.prompt.radius }}px;
             {% else %}
-                border-top-left-radius: {{ prompt_radius }}px;
-                border-top-right-radius: {{ prompt_radius }}px;
+                border-top-left-radius: {{ conf.prompt.radius }}px;
+                border-top-right-radius: {{ conf.prompt.radius }}px;
             {% endif %}
         }
 
         QWidget {
-            font: {{ font['prompts'] }};
-            color: {{ color['prompts.fg'] }};
-            background-color: {{ color['prompts.bg'] }};
+            font: {{ conf.fonts.prompts }};
+            color: {{ conf.colors.prompts.fg }};
+            background-color: {{ conf.colors.prompts.bg }};
+        }
+
+        QLineEdit {
+            border: {{ conf.colors.prompts.border }};
         }
 
         QTreeView {
-            selection-background-color: {{ color['prompts.selected.bg'] }};
+            selection-background-color: {{ conf.colors.prompts.selected.bg }};
+            border: {{ conf.colors.prompts.border }};
         }
 
-        QTreeView::item:selected, QTreeView::item:selected:hover {
-            background-color: {{ color['prompts.selected.bg'] }};
+        QTreeView::branch {
+            background-color: {{ conf.colors.prompts.bg }};
+        }
+
+        QTreeView::item:selected, QTreeView::item:selected:hover,
+        QTreeView::branch:selected {
+            background-color: {{ conf.colors.prompts.selected.bg }};
         }
     """
     update_geometry = pyqtSignal()
@@ -265,11 +289,11 @@ class PromptContainer(QWidget):
         self._layout = QVBoxLayout(self)
         self._layout.setContentsMargins(10, 10, 10, 10)
         self._win_id = win_id
-        self._prompt = None
+        self._prompt = None  # type: typing.Optional[_BasePrompt]
 
         self.setObjectName('PromptContainer')
         self.setAttribute(Qt.WA_StyledBackground, True)
-        style.set_register_stylesheet(self)
+        stylesheet.set_register(self)
 
         message.global_bridge.prompt_done.connect(self._on_prompt_done)
         prompt_queue.show_prompts.connect(self._on_show_prompts)
@@ -311,11 +335,10 @@ class PromptContainer(QWidget):
         log.prompt.debug("Displaying prompt {}".format(prompt))
         self._prompt = prompt
 
+        # If this question was interrupted, we already connected the signal
         if not question.interrupted:
-            # If this question was interrupted, we already connected the signal
             question.aborted.connect(
-                lambda: modeman.leave(self._win_id, prompt.KEY_MODE, 'aborted',
-                                      maybe=True))
+                functools.partial(self._on_aborted, prompt.KEY_MODE))
         modeman.enter(self._win_id, prompt.KEY_MODE, 'question asked')
 
         self.setSizePolicy(prompt.sizePolicy())
@@ -325,6 +348,15 @@ class PromptContainer(QWidget):
         prompt.setFocus()
         self.update_geometry.emit()
 
+    @pyqtSlot()
+    def _on_aborted(self, key_mode):
+        """Leave KEY_MODE whenever a prompt is aborted."""
+        try:
+            modeman.leave(self._win_id, key_mode, 'aborted', maybe=True)
+        except objreg.RegistryUnavailableError:
+            # window was deleted: ignore
+            pass
+
     @pyqtSlot(usertypes.KeyMode)
     def _on_prompt_done(self, key_mode):
         """Leave the prompt mode in this window if a question was answered."""
@@ -347,10 +379,10 @@ class PromptContainer(QWidget):
             widget.hide()
             widget.deleteLater()
 
-    @cmdutils.register(instance='prompt-container', hide=True, scope='window',
+    @cmdutils.register(instance='prompt-container', scope='window',
                        modes=[usertypes.KeyMode.prompt,
                               usertypes.KeyMode.yesno])
-    def prompt_accept(self, value=None):
+    def prompt_accept(self, value=None, *, save=False):
         """Accept the current prompt.
 
         //
@@ -361,33 +393,24 @@ class PromptContainer(QWidget):
         Args:
             value: If given, uses this value instead of the entered one.
                    For boolean prompts, "yes"/"no" are accepted as value.
+            save: Save the value to the config.
         """
+        assert self._prompt is not None
         question = self._prompt.question
+
         try:
-            done = self._prompt.accept(value)
+            done = self._prompt.accept(value, save=save)
         except Error as e:
-            raise cmdexc.CommandError(str(e))
+            raise cmdutils.CommandError(str(e))
+
         if done:
             message.global_bridge.prompt_done.emit(self._prompt.KEY_MODE)
             question.done()
 
-    @cmdutils.register(instance='prompt-container', hide=True, scope='window',
-                       modes=[usertypes.KeyMode.yesno],
-                       deprecated='Use :prompt-accept yes instead!')
-    def prompt_yes(self):
-        """Answer yes to a yes/no prompt."""
-        self.prompt_accept('yes')
-
-    @cmdutils.register(instance='prompt-container', hide=True, scope='window',
-                       modes=[usertypes.KeyMode.yesno],
-                       deprecated='Use :prompt-accept no instead!')
-    def prompt_no(self):
-        """Answer no to a yes/no prompt."""
-        self.prompt_accept('no')
-
-    @cmdutils.register(instance='prompt-container', hide=True, scope='window',
+    @cmdutils.register(instance='prompt-container', scope='window',
                        modes=[usertypes.KeyMode.prompt], maxsplit=0)
-    def prompt_open_download(self, cmdline: str = None):
+    def prompt_open_download(self, cmdline: str = None,
+                             pdfjs: bool = False) -> None:
         """Immediately open a download.
 
         If no specific command is given, this will use the system's default
@@ -398,13 +421,15 @@ class PromptContainer(QWidget):
                      is expanded to the temporary file name. If no `{}` is
                      present, the filename is automatically appended to the
                      cmdline.
+            pdfjs: Open the download via PDF.js.
         """
+        assert self._prompt is not None
         try:
-            self._prompt.download_open(cmdline)
+            self._prompt.download_open(cmdline, pdfjs=pdfjs)
         except UnsupportedOperationError:
             pass
 
-    @cmdutils.register(instance='prompt-container', hide=True, scope='window',
+    @cmdutils.register(instance='prompt-container', scope='window',
                        modes=[usertypes.KeyMode.prompt])
     @cmdutils.argument('which', choices=['next', 'prev'])
     def prompt_item_focus(self, which):
@@ -413,11 +438,34 @@ class PromptContainer(QWidget):
         Args:
             which: 'next', 'prev'
         """
+        assert self._prompt is not None
         try:
             self._prompt.item_focus(which)
         except UnsupportedOperationError:
             pass
 
+    @cmdutils.register(
+        instance='prompt-container', scope='window',
+        modes=[usertypes.KeyMode.prompt, usertypes.KeyMode.yesno])
+    def prompt_yank(self, sel=False):
+        """Yank URL to clipboard or primary selection.
+
+        Args:
+            sel: Use the primary selection instead of the clipboard.
+        """
+        assert self._prompt is not None
+        question = self._prompt.question
+        if question.url is None:
+            message.error('No URL found.')
+            return
+        if sel and utils.supports_selection():
+            target = 'primary selection'
+        else:
+            sel = False
+            target = 'clipboard'
+        utils.set_clipboard(question.url, sel)
+        message.info("Yanked to {}: {}".format(target, question.url))
+
 
 class LineEdit(QLineEdit):
 
@@ -427,7 +475,6 @@ class LineEdit(QLineEdit):
         super().__init__(parent)
         self.setStyleSheet("""
             QLineEdit {
-                border: 1px solid grey;
                 background-color: transparent;
             }
         """)
@@ -483,9 +530,8 @@ class _BasePrompt(QWidget):
         self._key_grid = QGridLayout()
         self._key_grid.setVerticalSpacing(0)
 
-        key_config = objreg.get('key-config')
-        # The bindings are all in the 'prompt' mode, even for yesno prompts
-        all_bindings = key_config.get_reverse_bindings_for('prompt')
+        all_bindings = config.key_instance.get_reverse_bindings_for(
+            self.KEY_MODE.name)
         labels = []
 
         for cmd, text in self._allowed_commands():
@@ -506,13 +552,23 @@ class _BasePrompt(QWidget):
             self._key_grid.addWidget(key_label, i, 0)
             self._key_grid.addWidget(text_label, i, 1)
 
+        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding)
+        self._key_grid.addItem(spacer, 0, 2)
+
         self._vbox.addLayout(self._key_grid)
 
-    def accept(self, value=None):
+    def _check_save_support(self, save):
+        if save:
+            raise UnsupportedOperationError("Saving answers is only possible "
+                                            "with yes/no prompts.")
+
+    def accept(self, value=None, save=False):
         raise NotImplementedError
 
-    def download_open(self, _cmdline):
+    def download_open(self, cmdline, pdfjs):
         """Open the download directly if this is a download prompt."""
+        utils.unused(cmdline)
+        utils.unused(pdfjs)
         raise UnsupportedOperationError
 
     def item_focus(self, _which):
@@ -538,7 +594,8 @@ class LineEditPrompt(_BasePrompt):
         self.setFocusProxy(self._lineedit)
         self._init_key_label()
 
-    def accept(self, value=None):
+    def accept(self, value=None, save=False):
+        self._check_save_support(save)
         text = value if value is not None else self._lineedit.text()
         self.question.answer = text
         return True
@@ -554,8 +611,7 @@ class FilenamePrompt(_BasePrompt):
     def __init__(self, question, parent=None):
         super().__init__(question, parent)
         self._init_texts(question)
-        self._init_fileview()
-        self._set_fileview_root(question.default)
+        self._init_key_label()
 
         self._lineedit = LineEdit(self)
         if question.default:
@@ -564,11 +620,15 @@ class FilenamePrompt(_BasePrompt):
         self._vbox.addWidget(self._lineedit)
 
         self.setFocusProxy(self._lineedit)
-        self._init_key_label()
 
-        if config.get('ui', 'prompt-filebrowser'):
+        self._init_fileview()
+        self._set_fileview_root(question.default)
+
+        if config.val.prompt.filebrowser:
             self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
 
+        self._to_complete = ''
+
     @pyqtSlot(str)
     def _set_fileview_root(self, path, *, tabbed=False):
         """Set the root path for the file display."""
@@ -577,6 +637,9 @@ class FilenamePrompt(_BasePrompt):
             separators += os.altsep
 
         dirname = os.path.dirname(path)
+        basename = os.path.basename(path)
+        if not tabbed:
+            self._to_complete = ''
 
         try:
             if not path:
@@ -590,6 +653,7 @@ class FilenamePrompt(_BasePrompt):
             elif os.path.isdir(dirname) and not tabbed:
                 # Input like /foo/ba -> show /foo contents
                 path = dirname
+                self._to_complete = basename
             else:
                 return
         except OSError:
@@ -607,7 +671,11 @@ class FilenamePrompt(_BasePrompt):
             index: The QModelIndex of the selected element.
             clicked: Whether the element was clicked.
         """
-        path = os.path.normpath(self._file_model.filePath(index))
+        if index == QModelIndex():
+            path = os.path.join(self._file_model.rootPath(), self._to_complete)
+        else:
+            path = os.path.normpath(self._file_model.filePath(index))
+
         if clicked:
             path += os.sep
         else:
@@ -628,7 +696,7 @@ class FilenamePrompt(_BasePrompt):
         self._file_view.setModel(self._file_model)
         self._file_view.clicked.connect(self._insert_path)
 
-        if config.get('ui', 'prompt-filebrowser'):
+        if config.val.prompt.filebrowser:
             self._vbox.addWidget(self._file_view)
         else:
             self._file_view.hide()
@@ -643,7 +711,8 @@ class FilenamePrompt(_BasePrompt):
         self._file_model.directoryLoaded.connect(
             lambda: self._file_model.sort(0))
 
-    def accept(self, value=None):
+    def accept(self, value=None, save=False):
+        self._check_save_support(save)
         text = value if value is not None else self._lineedit.text()
         text = downloads.transform_path(text)
         if text is None:
@@ -669,6 +738,7 @@ class FilenamePrompt(_BasePrompt):
         assert last_index.isValid()
 
         idx = selmodel.currentIndex()
+
         if not idx.isValid():
             # No item selected yet
             idx = last_index if which == 'prev' else first_index
@@ -682,10 +752,26 @@ class FilenamePrompt(_BasePrompt):
         if not idx.isValid():
             idx = last_index if which == 'prev' else first_index
 
+        idx = self._do_completion(idx, which)
+
         selmodel.setCurrentIndex(
-            idx, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
+            idx,
+            QItemSelectionModel.ClearAndSelect |  # type: ignore[arg-type]
+            QItemSelectionModel.Rows)
         self._insert_path(idx, clicked=False)
 
+    def _do_completion(self, idx, which):
+        filename = self._file_model.fileName(idx)
+        while not filename.startswith(self._to_complete) and idx.isValid():
+            if which == 'prev':
+                idx = self._file_view.indexAbove(idx)
+            else:
+                assert which == 'next', which
+                idx = self._file_view.indexBelow(idx)
+            filename = self._file_model.fileName(idx)
+
+        return idx
+
     def _allowed_commands(self):
         return [('prompt-accept', 'Accept'), ('leave-mode', 'Abort')]
 
@@ -696,17 +782,24 @@ class DownloadFilenamePrompt(FilenamePrompt):
 
     def __init__(self, question, parent=None):
         super().__init__(question, parent)
-        self._file_model.setFilter(QDir.AllDirs | QDir.Drives | QDir.NoDot)
+        self._file_model.setFilter(
+            QDir.AllDirs | QDir.Drives | QDir.NoDot)  # type: ignore[arg-type]
 
-    def accept(self, value=None):
-        done = super().accept(value)
+    def accept(self, value=None, save=False):
+        done = super().accept(value, save)
         answer = self.question.answer
         if answer is not None:
             self.question.answer = downloads.FileDownloadTarget(answer)
         return done
 
-    def download_open(self, cmdline):
-        self.question.answer = downloads.OpenFileDownloadTarget(cmdline)
+    def download_open(self, cmdline, pdfjs):
+        if pdfjs:
+            target = downloads.PDFJSDownloadTarget(
+            )  # type: downloads._DownloadTarget
+        else:
+            target = downloads.OpenFileDownloadTarget(cmdline)
+
+        self.question.answer = target
         self.question.done()
         message.global_bridge.prompt_done.emit(self.KEY_MODE)
 
@@ -715,6 +808,8 @@ class DownloadFilenamePrompt(FilenamePrompt):
             ('prompt-accept', 'Accept'),
             ('leave-mode', 'Abort'),
             ('prompt-open-download', "Open download"),
+            ('prompt-open-download --pdfjs', "Open download via PDF.js"),
+            ('prompt-yank', "Yank URL"),
         ]
         return cmds
 
@@ -745,14 +840,15 @@ class AuthenticationPrompt(_BasePrompt):
         assert not question.default, question.default
         self.setFocusProxy(self._user_lineedit)
 
-    def accept(self, value=None):
+    def accept(self, value=None, save=False):
+        self._check_save_support(save)
         if value is not None:
             if ':' not in value:
                 raise Error("Value needs to be in the format "
                             "username:password, but {} was given".format(
                                 value))
             username, password = value.split(':', maxsplit=1)
-            self.question.answer = AuthTuple(username, password)
+            self.question.answer = AuthInfo(username, password)
             return True
         elif self._user_lineedit.hasFocus():
             # Earlier, tab was bound to :prompt-accept, so to still support
@@ -760,8 +856,8 @@ class AuthenticationPrompt(_BasePrompt):
             self._password_lineedit.setFocus()
             return False
         else:
-            self.question.answer = AuthTuple(self._user_lineedit.text(),
-                                             self._password_lineedit.text())
+            self.question.answer = AuthInfo(self._user_lineedit.text(),
+                                            self._password_lineedit.text())
             return True
 
     def item_focus(self, which):
@@ -788,7 +884,14 @@ class YesNoPrompt(_BasePrompt):
         self._init_texts(question)
         self._init_key_label()
 
-    def accept(self, value=None):
+    def _check_save_support(self, save):
+        if save and self.question.option is None:
+            raise Error("No setting available to save the answer for this "
+                        "question.")
+
+    def accept(self, value=None, save=False):
+        self._check_save_support(save)
+
         if value is None:
             if self.question.default is None:
                 raise Error("No default value was set for this question!")
@@ -799,13 +902,30 @@ class YesNoPrompt(_BasePrompt):
             self.question.answer = False
         else:
             raise Error("Invalid value {} - expected yes/no!".format(value))
+
+        if save:
+            opt = config.instance.get_opt(self.question.option)
+            assert isinstance(opt.typ, configtypes.Bool)
+            pattern = urlmatch.UrlPattern(self.question.url)
+
+            try:
+                config.instance.set_obj(opt.name, self.question.answer,
+                                        pattern=pattern, save_yaml=True)
+            except configexc.Error as e:
+                raise Error(str(e))
+
         return True
 
     def _allowed_commands(self):
-        cmds = [
-            ('prompt-accept yes', "Yes"),
-            ('prompt-accept no', "No"),
-        ]
+        cmds = []
+
+        cmds.append(('prompt-accept yes', "Yes"))
+        if self.question.option is not None:
+            cmds.append(('prompt-accept --save yes', "Always"))
+
+        cmds.append(('prompt-accept no', "No"))
+        if self.question.option is not None:
+            cmds.append(('prompt-accept --save no', "Never"))
 
         if self.question.default is not None:
             assert self.question.default in [True, False]
@@ -813,6 +933,7 @@ class YesNoPrompt(_BasePrompt):
             cmds.append(('prompt-accept', "Use default ({})".format(default)))
 
         cmds.append(('leave-mode', "Abort"))
+        cmds.append(('prompt-yank', "Yank URL"))
         return cmds
 
 
@@ -825,7 +946,8 @@ class AlertPrompt(_BasePrompt):
         self._init_texts(question)
         self._init_key_label()
 
-    def accept(self, value=None):
+    def accept(self, value=None, save=False):
+        self._check_save_support(save)
         if value is not None:
             raise Error("No value is permitted with alert prompts!")
         # Simply mark prompt as done without setting self.question.answer
@@ -839,6 +961,5 @@ def init():
     """Initialize global prompt objects."""
     global prompt_queue
     prompt_queue = PromptQueue()
-    objreg.register('prompt-queue', prompt_queue)  # for commands
-    message.global_bridge.ask_question.connect(
+    message.global_bridge.ask_question.connect(  # type: ignore[call-arg]
         prompt_queue.ask_question, Qt.DirectConnection)
