@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -19,9 +19,11 @@
 
 """The tab widget used for TabbedBrowser from browser.py."""
 
-import collections
+import typing
 import functools
+import contextlib
 
+import attr
 from PyQt5.QtCore import (pyqtSignal, pyqtSlot, Qt, QSize, QRect, QPoint,
                           QTimer, QUrl)
 from PyQt5.QtWidgets import (QTabWidget, QTabBar, QSizePolicy, QCommonStyle,
@@ -30,12 +32,9 @@ from PyQt5.QtWidgets import (QTabWidget, QTabBar, QSizePolicy, QCommonStyle,
 from PyQt5.QtGui import QIcon, QPalette, QColor
 
 from qutebrowser.utils import qtutils, objreg, utils, usertypes, log
-from qutebrowser.config import config
-from qutebrowser.misc import objects
-
-
-PixelMetrics = usertypes.enum('PixelMetrics', ['icon_padding'],
-                              start=QStyle.PM_CustomBase, is_int=True)
+from qutebrowser.config import config, stylesheet
+from qutebrowser.misc import objects, debugcachestats
+from qutebrowser.browser import browsertab
 
 
 class TabWidget(QTabWidget):
@@ -46,40 +45,46 @@ class TabWidget(QTabWidget):
         tab_index_changed: Emitted when the current tab was changed.
                            arg 0: The index of the tab which is now focused.
                            arg 1: The total count of tabs.
+        new_tab_requested: Emitted when a new tab is requested.
     """
 
     tab_index_changed = pyqtSignal(int, int)
+    new_tab_requested = pyqtSignal('QUrl', bool, bool)
+
+    # Strings for controlling the mute/audible text
+    MUTE_STRING = '[M] '
+    AUDIBLE_STRING = '[A] '
 
     def __init__(self, win_id, parent=None):
         super().__init__(parent)
         bar = TabBar(win_id, self)
         self.setStyle(TabBarStyle())
         self.setTabBar(bar)
-        bar.tabCloseRequested.connect(self.tabCloseRequested)
+        bar.tabCloseRequested.connect(
+            self.tabCloseRequested)  # type: ignore[arg-type]
         bar.tabMoved.connect(functools.partial(
             QTimer.singleShot, 0, self.update_tab_titles))
         bar.currentChanged.connect(self._on_current_changed)
+        bar.new_tab_requested.connect(self._on_new_tab_requested)
         self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
         self.setDocumentMode(True)
         self.setElideMode(Qt.ElideRight)
         self.setUsesScrollButtons(True)
         bar.setDrawBase(False)
-        self.init_config()
-        objreg.get('config').changed.connect(self.init_config)
+        self._init_config()
+        config.instance.changed.connect(self._init_config)
 
     @config.change_filter('tabs')
-    def init_config(self):
+    def _init_config(self):
         """Initialize attributes based on the config."""
-        if self is None:  # pragma: no cover
-            # WORKAROUND for PyQt 5.2
-            return
         tabbar = self.tabBar()
-        self.setMovable(config.get('tabs', 'movable'))
+        self.setMovable(True)
         self.setTabsClosable(False)
-        position = config.get('tabs', 'position')
-        selection_behavior = config.get('tabs', 'select-on-remove')
+        position = config.val.tabs.position
+        selection_behavior = config.val.tabs.select_on_remove
         self.setTabPosition(position)
-        tabbar.vertical = position in [QTabWidget.West, QTabWidget.East]
+        tabbar.vertical = position in [  # type: ignore[attr-defined]
+            QTabWidget.West, QTabWidget.East]
         tabbar.setSelectionBehaviorOnRemove(selection_behavior)
         tabbar.refresh()
 
@@ -95,77 +100,93 @@ class TabWidget(QTabWidget):
         bar.update(bar.tabRect(idx))
 
     def set_tab_pinned(self, tab: QWidget,
-                       pinned: bool, *, loading: bool = False) -> None:
+                       pinned: bool) -> None:
         """Set the tab status as pinned.
 
         Args:
             tab: The tab to pin
             pinned: Pinned tab state to set.
-            loading: Whether to ignore current data state when
-                     counting pinned_count.
         """
-        bar = self.tabBar()
         idx = self.indexOf(tab)
-
-        # Only modify pinned_count if we had a change
-        # always modify pinned_count if we are loading
-        if tab.data.pinned != pinned or loading:
-            if pinned:
-                bar.pinned_count += 1
-            elif not pinned:
-                bar.pinned_count -= 1
-
-        bar.set_tab_data(idx, 'pinned', pinned)
         tab.data.pinned = pinned
+        self.update_tab_favicon(tab)
         self.update_tab_title(idx)
 
-        bar.refresh()
-
     def tab_indicator_color(self, idx):
         """Get the tab indicator color for the given index."""
         return self.tabBar().tab_indicator_color(idx)
 
     def set_page_title(self, idx, title):
         """Set the tab title user data."""
-        self.tabBar().set_tab_data(idx, 'page-title', title)
+        tabbar = self.tabBar()
+
+        if config.cache['tabs.tooltips']:
+            # always show only plain title in tooltips
+            tabbar.setTabToolTip(idx, title)
+
+        tabbar.set_tab_data(idx, 'page-title', title)
         self.update_tab_title(idx)
 
     def page_title(self, idx):
         """Get the tab title user data."""
         return self.tabBar().page_title(idx)
 
-    def update_tab_title(self, idx):
-        """Update the tab text for the given tab."""
+    def update_tab_title(self, idx, field=None):
+        """Update the tab text for the given tab.
+
+        Args:
+            idx: The tab index to update.
+            field: A field name which was updated. If given, the title
+                   is only set if the given field is in the template.
+        """
         tab = self.widget(idx)
+        if tab.data.pinned:
+            fmt = config.cache['tabs.title.format_pinned']
+        else:
+            fmt = config.cache['tabs.title.format']
+
+        if (field is not None and
+                (fmt is None or ('{' + field + '}') not in fmt)):
+            return
+
         fields = self.get_tab_fields(idx)
-        fields['title'] = fields['title'].replace('&', '&&')
+        fields['current_title'] = fields['current_title'].replace('&', '&&')
         fields['index'] = idx + 1
 
-        fmt = config.get('tabs', 'title-format')
-        fmt_pinned = config.get('tabs', 'title-format-pinned')
+        title = '' if fmt is None else fmt.format(**fields)
+        tabbar = self.tabBar()
 
-        if tab.data.pinned:
-            title = '' if fmt_pinned is None else fmt_pinned.format(**fields)
-        else:
-            title = '' if fmt is None else fmt.format(**fields)
-
-        self.tabBar().setTabText(idx, title)
+        # Only change the tab title if it changes, setting the tab title causes
+        # a size recalculation which is slow.
+        if tabbar.tabText(idx) != title:
+            tabbar.setTabText(idx, title)
 
     def get_tab_fields(self, idx):
         """Get the tab field data."""
         tab = self.widget(idx)
         if tab is None:
-            log.misc.debug("Got None-tab in get_tab_fields!")
+            log.misc.debug(  # type: ignore[unreachable]
+                "Got None-tab in get_tab_fields!")
 
         page_title = self.page_title(idx)
 
         fields = {}
         fields['id'] = tab.tab_id
-        fields['title'] = page_title
+        fields['current_title'] = page_title
         fields['title_sep'] = ' - ' if page_title else ''
         fields['perc_raw'] = tab.progress()
         fields['backend'] = objects.backend.name
-        fields['private'] = ' [Private Mode] ' if tab.private else ''
+        fields['private'] = ' [Private Mode] ' if tab.is_private else ''
+        try:
+            if tab.audio.is_muted():
+                fields['audio'] = TabWidget.MUTE_STRING
+            elif tab.audio.is_recently_audible():
+                fields['audio'] = TabWidget.AUDIBLE_STRING
+            else:
+                fields['audio'] = ''
+        except browsertab.WebTabError:
+            # Muting is only implemented with QtWebEngine
+            fields['audio'] = ''
 
         if tab.load_status() == usertypes.LoadStatus.loading:
             fields['perc'] = '[{}%] '.format(tab.progress())
@@ -173,9 +194,15 @@ class TabWidget(QTabWidget):
             fields['perc'] = ''
 
         try:
-            fields['host'] = self.tab_url(idx).host()
+            url = self.tab_url(idx)
         except qtutils.QtValueError:
             fields['host'] = ''
+            fields['current_url'] = ''
+            fields['protocol'] = ''
+        else:
+            fields['host'] = url.host()
+            fields['current_url'] = url.toDisplayString()
+            fields['protocol'] = url.scheme()
 
         y = tab.scroller.pos_perc()[1]
         if y is None:
@@ -190,10 +217,33 @@ class TabWidget(QTabWidget):
         fields['scroll_pos'] = scroll_pos
         return fields
 
-    def update_tab_titles(self, section='tabs', option='title-format'):
+    @contextlib.contextmanager
+    def _toggle_visibility(self):
+        """Toggle visibility while running.
+
+        Every single call to setTabText calls the size hinting functions for
+        every single tab, which are slow. Since we know we are updating all
+        the tab's titles, we can delay this processing by making the tab
+        non-visible. To avoid flickering, disable repaint updates whlie we
+        work.
+        """
+        bar = self.tabBar()
+        toggle = (self.count() > 10 and
+                  not bar.drag_in_progress and
+                  bar.isVisible())
+        if toggle:
+            bar.setUpdatesEnabled(False)
+            bar.setVisible(False)
+
+        yield
+
+        if toggle:
+            bar.setVisible(True)
+            bar.setUpdatesEnabled(True)
+
+    def update_tab_titles(self):
         """Update all texts."""
-        if section == 'tabs' and option in ['title-format',
-                                            'title-format-pinned']:
+        with self._toggle_visibility():
             for idx in range(self.count()):
                 self.update_tab_title(idx)
 
@@ -225,7 +275,6 @@ class TabWidget(QTabWidget):
             The index of the newly added tab.
         """
         if text_or_empty is None:
-            icon = None
             text = icon_or_text
             new_idx = super().addTab(page, '')
         else:
@@ -255,7 +304,6 @@ class TabWidget(QTabWidget):
             The index of the newly added tab.
         """
         if text_or_empty is None:
-            icon = None
             text = icon_or_text
             new_idx = super().insertTab(idx, page, '')
         else:
@@ -271,6 +319,11 @@ class TabWidget(QTabWidget):
         self.tabBar().on_current_changed()
         self.tab_index_changed.emit(index, self.count())
 
+    @pyqtSlot()
+    def _on_new_tab_requested(self):
+        """Open a new tab."""
+        self.new_tab_requested.emit(config.val.url.default_page, False, False)
+
     def tab_url(self, idx):
         """Get the URL of the tab at the given index.
 
@@ -279,13 +332,38 @@ class TabWidget(QTabWidget):
         """
         tab = self.widget(idx)
         if tab is None:
-            url = QUrl()
+            url = QUrl()  # type: ignore[unreachable]
         else:
             url = tab.url()
         # It's possible for url to be invalid, but the caller will handle that.
         qtutils.ensure_valid(url)
         return url
 
+    def update_tab_favicon(self, tab: QWidget) -> None:
+        """Update favicon of the given tab."""
+        idx = self.indexOf(tab)
+
+        if tab.data.should_show_icon():
+            self.setTabIcon(idx, tab.icon())
+            if config.val.tabs.tabs_are_windows:
+                self.window().setWindowIcon(tab.icon())
+        else:
+            self.setTabIcon(idx, QIcon())
+            if config.val.tabs.tabs_are_windows:
+                self.window().setWindowIcon(self.window().windowIcon())
+
+    def setTabIcon(self, idx: int, icon: QIcon) -> None:
+        """Always show tab icons for pinned tabs in some circumstances."""
+        tab = typing.cast(typing.Optional[browsertab.AbstractTab],
+                          self.widget(idx))
+        if (icon.isNull() and
+                config.cache['tabs.favicons.show'] != 'never' and
+                config.cache['tabs.pinned.shrink'] and
+                not self.tabBar().vertical and
+                tab is not None and tab.data.pinned):
+            icon = self.style().standardIcon(QStyle.SP_FileIcon)
+        super().setTabIcon(idx, icon)
+
 
 class TabBar(QTabBar):
 
@@ -300,30 +378,40 @@ class TabBar(QTabBar):
     Attributes:
         vertical: When the tab bar is currently vertical.
         win_id: The window ID this TabBar belongs to.
+
+    Signals:
+        new_tab_requested: Emitted when a new tab is requested.
     """
 
+    STYLESHEET = """
+        TabBar {
+            font: {{ conf.fonts.tabs.unselected }};
+            background-color: {{ conf.colors.tabs.bar.bg }};
+        }
+
+        TabBar::tab:selected {
+            font: {{ conf.fonts.tabs.selected }};
+        }
+    """
+
+    new_tab_requested = pyqtSignal()
+
     def __init__(self, win_id, parent=None):
         super().__init__(parent)
         self._win_id = win_id
         self.setStyle(TabBarStyle())
-        self.set_font()
-        config_obj = objreg.get('config')
-        config_obj.changed.connect(self.set_font)
-        config_obj.changed.connect(self.set_icon_size)
         self.vertical = False
         self._auto_hide_timer = QTimer()
         self._auto_hide_timer.setSingleShot(True)
-        self._auto_hide_timer.setInterval(
-            config.get('tabs', 'show-switching-delay'))
         self._auto_hide_timer.timeout.connect(self.maybe_hide)
+        self._on_show_switching_delay_changed()
         self.setAutoFillBackground(True)
-        self.set_colors()
-        self.pinned_count = 0
-        config_obj.changed.connect(self.set_colors)
+        self.drag_in_progress = False
+        stylesheet.set_register(self)
+        self.ensurePolished()
+        config.instance.changed.connect(self._on_config_changed)
+        self._set_icon_size()
         QTimer.singleShot(0, self.maybe_hide)
-        config_obj.changed.connect(self.on_tab_colors_changed)
-        config_obj.changed.connect(self.on_show_switching_delay_changed)
-        config_obj.changed.connect(self.tabs_show)
 
     def __repr__(self):
         return utils.get_repr(self, count=self.count())
@@ -332,29 +420,47 @@ class TabBar(QTabBar):
         """Get the current tab object."""
         return self.parent().currentWidget()
 
-    @config.change_filter('tabs', 'show')
-    def tabs_show(self):
-        """Hide or show tab bar if needed when tabs->show got changed."""
-        self.maybe_hide()
+    @pyqtSlot(str)
+    def _on_config_changed(self, option: str) -> None:
+        if option.startswith('fonts.tabs.'):
+            self.ensurePolished()
+            self._set_icon_size()
+        elif option == 'tabs.favicons.scale':
+            self._set_icon_size()
+        elif option == 'tabs.show_switching_delay':
+            self._on_show_switching_delay_changed()
+        elif option == 'tabs.show':
+            self.maybe_hide()
 
-    @config.change_filter('tabs', 'show-switching-delay')
-    def on_show_switching_delay_changed(self):
-        """Set timer interval when tabs->show-switching-delay got changed."""
-        self._auto_hide_timer.setInterval(
-            config.get('tabs', 'show-switching-delay'))
+        if option.startswith('colors.tabs.'):
+            self.update()
+
+        # Clear tab size caches when appropriate
+        if option in ["tabs.indicator.padding",
+                      "tabs.padding",
+                      "tabs.indicator.width",
+                      "tabs.min_width",
+                      "tabs.pinned.shrink",
+                      "fonts.tabs.selected",
+                      "fonts.tabs.unselected"]:
+            self._minimum_tab_size_hint_helper.cache_clear()
+            self._minimum_tab_height.cache_clear()
+
+    def _on_show_switching_delay_changed(self):
+        """Set timer interval when tabs.show_switching_delay got changed."""
+        self._auto_hide_timer.setInterval(config.val.tabs.show_switching_delay)
 
     def on_current_changed(self):
         """Show tab bar when current tab got changed."""
         self.maybe_hide()  # for fullscreen tabs
-        show = config.get('tabs', 'show')
-        if show == 'switching':
+        if config.val.tabs.show == 'switching':
             self.show()
             self._auto_hide_timer.start()
 
     @pyqtSlot()
     def maybe_hide(self):
         """Hide the tab bar if needed."""
-        show = config.get('tabs', 'show')
+        show = config.val.tabs.show
         tab = self._current_tab()
         if (show in ['never', 'switching'] or
                 (show == 'multiple' and self.count() == 1) or
@@ -409,74 +515,124 @@ class TabBar(QTabBar):
         # code sets layoutDirty so it actually relayouts the tabs.
         self.setIconSize(self.iconSize())
 
-    @config.change_filter('fonts', 'tabbar')
-    def set_font(self):
-        """Set the tab bar font."""
-        self.setFont(config.get('fonts', 'tabbar'))
-        self.set_icon_size()
-
-    @config.change_filter('tabs', 'favicon-scale')
-    def set_icon_size(self):
+    def _set_icon_size(self):
         """Set the tab bar favicon size."""
         size = self.fontMetrics().height() - 2
-        size *= config.get('tabs', 'favicon-scale')
+        size = int(size * config.val.tabs.favicons.scale)
         self.setIconSize(QSize(size, size))
 
-    @config.change_filter('colors', 'tabs.bg.bar')
-    def set_colors(self):
-        """Set the tab bar colors."""
-        p = self.palette()
-        p.setColor(QPalette.Window, config.get('colors', 'tabs.bg.bar'))
-        self.setPalette(p)
-
-    @pyqtSlot(str, str)
-    def on_tab_colors_changed(self, section, option):
-        """Set the tab colors."""
-        if section == 'colors' and option.startswith('tabs.'):
-            self.update()
+    def mouseReleaseEvent(self, e):
+        """Override mouseReleaseEvent to know when drags stop."""
+        self.drag_in_progress = False
+        super().mouseReleaseEvent(e)
 
     def mousePressEvent(self, e):
-        """Override mousePressEvent to close tabs if configured."""
-        button = config.get('tabs', 'close-mouse-button')
+        """Override mousePressEvent to close tabs if configured.
+
+        Also keep track of if we are currently in a drag."""
+        self.drag_in_progress = True
+        button = config.val.tabs.close_mouse_button
         if (e.button() == Qt.RightButton and button == 'right' or
                 e.button() == Qt.MiddleButton and button == 'middle'):
             e.accept()
             idx = self.tabAt(e.pos())
             if idx == -1:
-                idx = self.currentIndex()
+                action = config.val.tabs.close_mouse_button_on_bar
+                if action == 'ignore':
+                    return
+                elif action == 'new-tab':
+                    self.new_tab_requested.emit()
+                    return
+                elif action == 'close-current':
+                    idx = self.currentIndex()
+                elif action == 'close-last':
+                    idx = self.count() - 1
             self.tabCloseRequested.emit(idx)
             return
         super().mousePressEvent(e)
 
-    def minimumTabSizeHint(self, index):
+    def minimumTabSizeHint(self, index: int, ellipsis: bool = True) -> QSize:
         """Set the minimum tab size to indicator/icon/... text.
 
         Args:
             index: The index of the tab to get a size hint for.
-
+            ellipsis: Whether to use ellipsis to calculate width
+                      instead of the tab's text.
+                      Forced to False for pinned tabs.
         Return:
-            A QSize.
+            A QSize of the smallest tab size we can make.
         """
         icon = self.tabIcon(index)
-        padding = config.get('tabs', 'padding')
-        padding_h = padding.left + padding.right
-        padding_v = padding.top + padding.bottom
         if icon.isNull():
-            icon_size = QSize(0, 0)
+            icon_width = 0
         else:
-            extent = self.style().pixelMetric(QStyle.PM_TabBarIconSize, None,
-                                              self)
-            icon_size = icon.actualSize(QSize(extent, extent))
-            padding_h += self.style().pixelMetric(
-                PixelMetrics.icon_padding, None, self)
-        indicator_width = config.get('tabs', 'indicator-width')
-        height = self.fontMetrics().height() + padding_v
-        width = (self.fontMetrics().width('\u2026') + icon_size.width() +
+            icon_width = min(
+                icon.actualSize(self.iconSize()).width(),
+                self.iconSize().width()) + TabBarStyle.ICON_PADDING
+
+        pinned = self._tab_pinned(index)
+        if not self.vertical and pinned and config.val.tabs.pinned.shrink:
+            # Never consider ellipsis an option for horizontal pinned tabs
+            ellipsis = False
+        return self._minimum_tab_size_hint_helper(self.tabText(index),
+                                                  icon_width, ellipsis,
+                                                  pinned)
+
+    @debugcachestats.register(name='tab width cache')
+    @functools.lru_cache(maxsize=2**9)
+    def _minimum_tab_size_hint_helper(self, tab_text: str,
+                                      icon_width: int,
+                                      ellipsis: bool, pinned: bool) -> QSize:
+        """Helper function to cache tab results.
+
+        Config values accessed in here should be added to _on_config_changed to
+        ensure cache is flushed when needed.
+        """
+        text = '\u2026' if ellipsis else tab_text
+        # Don't ever shorten if text is shorter than the ellipsis
+
+        def _text_to_width(text):
+            # Calculate text width taking into account qt mnemonics
+            return self.fontMetrics().size(Qt.TextShowMnemonic, text).width()
+        text_width = min(_text_to_width(text),
+                         _text_to_width(tab_text))
+        padding = config.cache['tabs.padding']
+        indicator_width = config.cache['tabs.indicator.width']
+        indicator_padding = config.cache['tabs.indicator.padding']
+        padding_h = padding.left + padding.right
+
+        # Only add padding if indicator exists
+        if indicator_width != 0:
+            padding_h += indicator_padding.left + indicator_padding.right
+        height = self._minimum_tab_height()
+        width = (text_width + icon_width +
                  padding_h + indicator_width)
+        min_width = config.cache['tabs.min_width']
+        if (not self.vertical and min_width > 0 and
+                not pinned or not config.cache['tabs.pinned.shrink']):
+            width = max(min_width, width)
         return QSize(width, height)
 
-    def tabSizeHint(self, index):
-        """Override tabSizeHint so all tabs are the same size.
+    @functools.lru_cache(maxsize=1)
+    def _minimum_tab_height(self):
+        padding = config.cache['tabs.padding']
+        return self.fontMetrics().height() + padding.top + padding.bottom
+
+    def _tab_pinned(self, index: int) -> bool:
+        """Return True if tab is pinned."""
+        if not 0 <= index < self.count():
+            raise IndexError("Tab index ({}) out of range ({})!".format(
+                index, self.count()))
+
+        widget = self.parent().widget(index)
+        if widget is None:
+            # This could happen when Qt calls tabSizeHint while initializing
+            # tabs.
+            return False
+        return widget.data.pinned
+
+    def tabSizeHint(self, index: int) -> QSize:
+        """Override tabSizeHint to customize qb's tab size.
 
         https://wiki.python.org/moin/PyQt/Customising%20tab%20bars
 
@@ -486,10 +642,15 @@ class TabBar(QTabBar):
         Return:
             A QSize.
         """
-        minimum_size = self.minimumTabSizeHint(index)
-        height = minimum_size.height()
+        if self.count() == 0:
+            # This happens on startup on macOS.
+            # We return it directly rather than setting `size' because we don't
+            # want to ensure it's valid in this special case.
+            return QSize()
+
+        height = self._minimum_tab_height()
         if self.vertical:
-            confwidth = str(config.get('tabs', 'width'))
+            confwidth = str(config.cache['tabs.width'])
             if confwidth.endswith('%'):
                 main_window = objreg.get('main-window', scope='window',
                                          window=self._win_id)
@@ -497,90 +658,50 @@ class TabBar(QTabBar):
                 width = main_window.width() * perc / 100
             else:
                 width = int(confwidth)
-            size = QSize(max(minimum_size.width(), width), height)
-        elif self.count() == 0:
-            # This happens on startup on OS X.
-            # We return it directly rather than setting `size' because we don't
-            # want to ensure it's valid in this special case.
-            return QSize()
+            size = QSize(width, height)
         else:
-            tab_width_pinned_conf = config.get('tabs', 'pinned-width')
-
-            try:
-                pinned = self.tab_data(index, 'pinned')
-            except KeyError:
-                pinned = False
-
-            no_pinned_count = self.count() - self.pinned_count
-            pinned_width = tab_width_pinned_conf * self.pinned_count
-            # Prevent any tabs from being smaller than the min size
-            no_pinned_width = max(self.width() - pinned_width,
-                                  minimum_size.width() * no_pinned_count)
-
-            if pinned:
-                width = tab_width_pinned_conf
+            if config.cache['tabs.pinned.shrink'] and self._tab_pinned(index):
+                # Give pinned tabs the minimum size they need to display their
+                # titles, let Qt handle scaling it down if we get too small.
+                width = self.minimumTabSizeHint(index, ellipsis=False).width()
             else:
-
-                # Tabs should attempt to occupy the whole window width. If
-                # there are pinned tabs their size will be subtracted from the
-                # total window width.  During shutdown the self.count goes
-                # down, but the self.pinned_count not - this generates some odd
-                # behavior. To avoid this we compare self.count against
-                # self.pinned_count. If we end up having too little space, we
-                # set the minimum size below.
-                if self.pinned_count > 0 and no_pinned_count > 0:
-                    width = no_pinned_width / no_pinned_count
-                else:
-                    width = self.width() / self.count()
-
-            # If no_pinned_width is not divisible by no_pinned_count, add a
-            # pixel to some tabs so that there is no ugly leftover space.
-            if (no_pinned_count > 0 and
-                    index < no_pinned_width % no_pinned_count):
-                width += 1
-
-            # If we don't have enough space, we return the minimum size so we
-            # get scroll buttons as soon as needed.
-            width = max(width, minimum_size.width())
-
+                # Request as much space as possible so we fill the tabbar, let
+                # Qt shrink us down. If for some reason (tests, bugs)
+                # self.width() gives 0, use a sane min of 10 px
+                width = max(self.width(), 10)
+                max_width = config.cache['tabs.max_width']
+                if max_width > 0:
+                    width = min(max_width, width)
             size = QSize(width, height)
         qtutils.ensure_valid(size)
         return size
 
-    def paintEvent(self, _e):
+    def paintEvent(self, event):
         """Override paintEvent to draw the tabs like we want to."""
-        # pylint: disable=bad-config-call
-        # WORKAROUND for https://bitbucket.org/logilab/astroid/issue/104
         p = QStylePainter(self)
         selected = self.currentIndex()
         for idx in range(self.count()):
+            if not event.region().intersects(self.tabRect(idx)):
+                # Don't repaint if we are outside the requested region
+                continue
+
             tab = QStyleOptionTab()
             self.initStyleOption(tab, idx)
 
-            bg_parts = ['tabs', 'bg']
-            fg_parts = ['tabs', 'fg']
+            setting = 'colors.tabs'
+            if self._tab_pinned(idx):
+                setting += '.pinned'
             if idx == selected:
-                bg_parts.append('selected')
-                fg_parts.append('selected')
+                setting += '.selected'
+            setting += '.odd' if (idx + 1) % 2 else '.even'
 
-            if idx % 2:
-                bg_parts.append('odd')
-                fg_parts.append('odd')
-            else:
-                bg_parts.append('even')
-                fg_parts.append('even')
-
-            bg_color = config.get('colors', '.'.join(bg_parts))
-            fg_color = config.get('colors', '.'.join(fg_parts))
-            tab.palette.setColor(QPalette.Window, bg_color)
-            tab.palette.setColor(QPalette.WindowText, fg_color)
+            tab.palette.setColor(QPalette.Window,
+                                 config.cache[setting + '.bg'])
+            tab.palette.setColor(QPalette.WindowText,
+                                 config.cache[setting + '.fg'])
 
             indicator_color = self.tab_indicator_color(idx)
             tab.palette.setColor(QPalette.Base, indicator_color)
-            if tab.rect.right() < 0 or tab.rect.left() > self.width():
-                # Don't bother drawing a tab if the entire tab is outside of
-                # the visible tab bar.
-                continue
             p.drawControl(QStyle.CE_TabBarTab, tab)
 
     def tabInserted(self, idx):
@@ -599,7 +720,7 @@ class TabBar(QTabBar):
         Args:
             e: The QWheelEvent
         """
-        if config.get('tabs', 'mousewheel-tab-switching'):
+        if config.val.tabs.mousewheel_switching:
             super().wheelEvent(e)
         else:
             tabbed_browser = objreg.get('tabbed-browser', scope='window',
@@ -607,8 +728,17 @@ class TabBar(QTabBar):
             tabbed_browser.wheelEvent(e)
 
 
-# Used by TabBarStyle._tab_layout().
-Layouts = collections.namedtuple('Layouts', ['text', 'icon', 'indicator'])
+@attr.s
+class Layouts:
+
+    """Layout information for tab.
+
+    Used by TabBarStyle._tab_layout().
+    """
+
+    text = attr.ib()
+    icon = attr.ib()
+    indicator = attr.ib()
 
 
 class TabBarStyle(QCommonStyle):
@@ -629,6 +759,8 @@ class TabBarStyle(QCommonStyle):
     https://code.google.com/p/makehuman/source/browse/trunk/makehuman/lib/qtgui.py
     """
 
+    ICON_PADDING = 4
+
     def __init__(self):
         """Initialize all functions we're not overriding.
 
@@ -671,7 +803,7 @@ class TabBarStyle(QCommonStyle):
         icon_state = (QIcon.On if opt.state & QStyle.State_Selected
                       else QIcon.Off)
         icon = opt.icon.pixmap(opt.iconSize, icon_mode, icon_state)
-        p.drawPixmap(layouts.icon.x(), layouts.icon.y(), icon)
+        self._style.drawItemPixmap(p, layouts.icon, Qt.AlignCenter, icon)
 
     def drawControl(self, element, opt, p, widget=None):
         """Override drawControl to draw odd tabs in a different color.
@@ -709,11 +841,15 @@ class TabBarStyle(QCommonStyle):
         elif element == QStyle.CE_TabBarTabLabel:
             if not opt.icon.isNull() and layouts.icon.isValid():
                 self._draw_icon(layouts, opt, p)
-            alignment = (config.get('tabs', 'title-alignment') |
+            alignment = (config.cache['tabs.title.alignment'] |
                          Qt.AlignVCenter | Qt.TextHideMnemonic)
-            self._style.drawItemText(p, layouts.text, alignment, opt.palette,
-                                     opt.state & QStyle.State_Enabled,
-                                     opt.text, QPalette.WindowText)
+            self._style.drawItemText(p,
+                                     layouts.text,
+                                     int(alignment),
+                                     opt.palette,
+                                     bool(opt.state & QStyle.State_Enabled),
+                                     opt.text,
+                                     QPalette.WindowText)
         else:
             raise ValueError("Invalid element {!r}".format(element))
 
@@ -734,8 +870,6 @@ class TabBarStyle(QCommonStyle):
                       QStyle.PM_TabBarTabVSpace,
                       QStyle.PM_TabBarScrollButtonWidth]:
             return 0
-        elif metric == PixelMetrics.icon_padding:
-            return 4
         else:
             return self._style.pixelMetric(metric, option, widget)
 
@@ -756,24 +890,17 @@ class TabBarStyle(QCommonStyle):
                 log.misc.warning("Could not get layouts for tab!")
                 return QRect()
             return layouts.text
-        elif sr == QStyle.SE_TabWidgetTabBar:
+        elif sr in [QStyle.SE_TabWidgetTabBar,
+                    QStyle.SE_TabBarScrollLeftButton]:
+            # Handling SE_TabBarScrollLeftButton so the left scroll button is
+            # aligned properly. Otherwise, empty space will be shown after the
+            # last tab even though the button width is set to 0
+            #
             # Need to use super() because we also use super() to render
             # element in drawControl(); otherwise, we may get bit by
             # style differences...
-            rct = super().subElementRect(sr, opt, widget)
-            return rct
+            return super().subElementRect(sr, opt, widget)
         else:
-            try:
-                # We need this so the left scroll button is aligned properly.
-                # Otherwise, empty space will be shown after the last tab even
-                # though the button width is set to 0
-                #
-                # QStyle.SE_TabBarScrollLeftButton was added in Qt 5.7
-                if sr == QStyle.SE_TabBarScrollLeftButton:
-                    return super().subElementRect(sr, opt, widget)
-            except AttributeError:
-                pass
-
             return self._style.subElementRect(sr, opt, widget)
 
     def _tab_layout(self, opt):
@@ -787,10 +914,10 @@ class TabBarStyle(QCommonStyle):
             opt: QStyleOptionTab
 
         Return:
-            A Layout namedtuple with two QRects.
+            A Layout object with two QRects.
         """
-        padding = config.get('tabs', 'padding')
-        indicator_padding = config.get('tabs', 'indicator-padding')
+        padding = config.cache['tabs.padding']
+        indicator_padding = config.cache['tabs.indicator.padding']
 
         text_rect = QRect(opt.rect)
         if not text_rect.isValid():
@@ -801,7 +928,7 @@ class TabBarStyle(QCommonStyle):
         text_rect.adjust(padding.left, padding.top, -padding.right,
                          -padding.bottom)
 
-        indicator_width = config.get('tabs', 'indicator-width')
+        indicator_width = config.cache['tabs.indicator.width']
         if indicator_width == 0:
             indicator_rect = QRect()
         else:
@@ -818,8 +945,8 @@ class TabBarStyle(QCommonStyle):
 
         icon_rect = self._get_icon_rect(opt, text_rect)
         if icon_rect.isValid():
-            icon_padding = self.pixelMetric(PixelMetrics.icon_padding, opt)
-            text_rect.adjust(icon_rect.width() + icon_padding, 0, 0, 0)
+            text_rect.adjust(
+                icon_rect.width() + TabBarStyle.ICON_PADDING, 0, 0, 0)
 
         text_rect = self._style.visualRect(opt.direction, opt.rect, text_rect)
         return Layouts(text=text_rect, icon=icon_rect,
@@ -844,17 +971,17 @@ class TabBarStyle(QCommonStyle):
         icon_state = (QIcon.On if opt.state & QStyle.State_Selected
                       else QIcon.Off)
         # reserve space for favicon when tab bar is vertical (issue #1968)
-        position = config.get('tabs', 'position')
-        if (opt.icon.isNull() and
-                position in [QTabWidget.East, QTabWidget.West] and
-                config.get('tabs', 'show-favicons')):
+        position = config.cache['tabs.position']
+        if (position in [QTabWidget.East, QTabWidget.West] and
+                config.cache['tabs.favicons.show'] != 'never'):
             tab_icon_size = icon_size
         else:
             actual_size = opt.icon.actualSize(icon_size, icon_mode, icon_state)
             tab_icon_size = QSize(
                 min(actual_size.width(), icon_size.width()),
                 min(actual_size.height(), icon_size.height()))
-        icon_top = text_rect.center().y() + 1 - tab_icon_size.height() / 2
+
+        icon_top = text_rect.center().y() + 1 - tab_icon_size.height() // 2
         icon_rect = QRect(QPoint(text_rect.left(), icon_top), tab_icon_size)
         icon_rect = self._style.visualRect(opt.direction, opt.rect, icon_rect)
         return icon_rect
