@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -27,18 +27,29 @@ import getpass
 import fnmatch
 import traceback
 import datetime
+import enum
+import typing
 
 import pkg_resources
 from PyQt5.QtCore import pyqtSlot, Qt, QSize
 from PyQt5.QtWidgets import (QDialog, QLabel, QTextEdit, QPushButton,
                              QVBoxLayout, QHBoxLayout, QCheckBox,
-                             QDialogButtonBox, QMessageBox, QApplication)
+                             QDialogButtonBox, QApplication, QMessageBox)
 
 import qutebrowser
-from qutebrowser.utils import version, log, utils, objreg, usertypes
+from qutebrowser.utils import version, log, utils
 from qutebrowser.misc import (miscwidgets, autoupdate, msgbox, httpclient,
-                              pastebin, objects)
-from qutebrowser.config import config
+                              pastebin)
+from qutebrowser.config import config, configfiles
+from qutebrowser.browser import history
+
+
+class Result(enum.IntEnum):
+
+    """The result code returned by the crash dialog."""
+
+    restore = QDialog.Accepted + 1
+    no_restore = QDialog.Accepted + 2
 
 
 def parse_fatal_stacktrace(text):
@@ -52,52 +63,22 @@ def parse_fatal_stacktrace(text):
         element being the first stacktrace frame.
     """
     lines = [
-        r'Fatal Python error: (.*)',
+        r'(?P<type>Fatal Python error|Windows fatal exception): (?P<msg>.*)',
         r' *',
         r'(Current )?[Tt]hread [^ ]* \(most recent call first\): *',
-        r'  File ".*", line \d+ in (.*)',
+        r'  File ".*", line \d+ in (?P<func>.*)',
     ]
-    m = re.match('\n'.join(lines), text)
+    m = re.search('\n'.join(lines), text)
     if m is None:
         # We got some invalid text.
         return ('', '')
     else:
-        return (m.group(1), m.group(3))
-
-
-def get_fatal_crash_dialog(debug, data):
-    """Get a fatal crash dialog based on a crash log.
-
-    If the crash is a segfault in qt_mainloop and we're on an old Qt version
-    this is a simple error dialog which lets the user know they should upgrade
-    if possible.
-
-    If it's anything else, it's a normal FatalCrashDialog with the possibility
-    to report the crash.
-
-    Args:
-        debug: Whether the debug flag (--debug) was given.
-        data: The crash log data.
-    """
-    ignored_frames = ['qt_mainloop', 'paintEvent']
-    errtype, frame = parse_fatal_stacktrace(data)
-
-    if (errtype == 'Segmentation fault' and
-            frame in ignored_frames and
-            objects.backend == usertypes.Backend.QtWebKit):
-        title = "qutebrowser was restarted after a fatal crash!"
-        text = ("<b>qutebrowser was restarted after a fatal crash!</b><br/>"
-                "Unfortunately, this crash occurred in Qt (the library "
-                "qutebrowser uses), and QtWebKit (the current backend) is not "
-                "maintained anymore.<br/><br/>Since I can't do much about "
-                "those crashes I disabled the crash reporter for this case, "
-                "but this will likely be resolved in the future with the new "
-                "QtWebEngine backend.")
-        box = QMessageBox(QMessageBox.Critical, title, text, QMessageBox.Ok)
-        box.setAttribute(Qt.WA_DeleteOnClose)
-        return box
-    else:
-        return FatalCrashDialog(debug, data)
+        msg = m.group('msg')
+        typ = m.group('type')
+        func = m.group('func')
+        if typ == 'Windows fatal exception':
+            msg = 'Windows ' + msg
+        return msg, func
 
 
 def _get_environment_vars():
@@ -138,59 +119,44 @@ class _CrashDialog(QDialog):
         super().__init__(parent)
         # We don't set WA_DeleteOnClose here as on an exception, we'll get
         # closed anyways, and it only could have unintended side-effects.
-        self._crash_info = []
+        self._crash_info = []  # type: typing.List[typing.Tuple[str, str]]
         self._btn_box = None
-        self._btn_report = None
-        self._btn_cancel = None
-        self._lbl = None
         self._paste_text = None
         self.setWindowTitle("Whoops!")
         self.resize(QSize(640, 600))
         self._vbox = QVBoxLayout(self)
+
         http_client = httpclient.HTTPClient()
         self._paste_client = pastebin.PastebinClient(http_client, self)
         self._pypi_client = autoupdate.PyPIVersionClient(self)
+        self._paste_client.success.connect(self.on_paste_success)
+        self._paste_client.error.connect(self.show_error)
+
         self._init_text()
 
-        contact = QLabel("I'd like to be able to follow up with you, to keep "
-                         "you posted on the status of this crash and get more "
-                         "information if I need it - how can I contact you?",
-                         wordWrap=True)
-        self._vbox.addWidget(contact)
-        self._contact = QTextEdit(tabChangesFocus=True, acceptRichText=False)
-        try:
-            state = objreg.get('state-config')
-            try:
-                self._contact.setPlainText(state['general']['contact-info'])
-            except KeyError:
-                self._contact.setPlaceholderText("Mail or IRC nickname")
-        except Exception:
-            log.misc.exception("Failed to get contact information!")
-            self._contact.setPlaceholderText("Mail or IRC nickname")
-        self._vbox.addWidget(self._contact, 2)
+        self._init_contact_input()
 
         info = QLabel("What were you doing when this crash/bug happened?")
         self._vbox.addWidget(info)
-        self._info = QTextEdit(tabChangesFocus=True, acceptRichText=False)
+        self._info = QTextEdit()
+        self._info.setTabChangesFocus(True)
+        self._info.setAcceptRichText(False)
         self._info.setPlaceholderText("- Opened http://www.example.com/\n"
                                       "- Switched tabs\n"
                                       "- etc...")
         self._vbox.addWidget(self._info, 5)
 
         self._vbox.addSpacing(15)
-        self._debug_log = QTextEdit(tabChangesFocus=True, acceptRichText=False,
-                                    lineWrapMode=QTextEdit.NoWrap)
+        self._debug_log = QTextEdit()
+        self._debug_log.setTabChangesFocus(True)
+        self._debug_log.setAcceptRichText(False)
+        self._debug_log.setLineWrapMode(QTextEdit.NoWrap)
         self._debug_log.hide()
-        info = QLabel("<i>You can edit the log below to remove sensitive "
-                      "information.</i>", wordWrap=True)
-        info.hide()
         self._fold = miscwidgets.DetailFold("Show log", self)
         self._fold.toggled.connect(self._debug_log.setVisible)
-        self._fold.toggled.connect(info.setVisible)
         if debug:
             self._fold.toggle()
         self._vbox.addWidget(self._fold)
-        self._vbox.addWidget(info)
         self._vbox.addWidget(self._debug_log, 10)
         self._vbox.addSpacing(15)
 
@@ -198,40 +164,73 @@ class _CrashDialog(QDialog):
         self._init_info_text()
         self._init_buttons()
 
+    def keyPressEvent(self, e):
+        """Prevent closing :report dialogs when pressing <Escape>."""
+        if config.val.input.escape_quits_reporter or e.key() != Qt.Key_Escape:
+            super().keyPressEvent(e)
+
     def __repr__(self):
         return utils.get_repr(self)
 
+    def _init_contact_input(self):
+        """Initialize the widget asking for contact info."""
+        contact = QLabel("I'd like to be able to follow up with you, to keep "
+                         "you posted on the status of this crash and get more "
+                         "information if I need it - how can I contact you?")
+        contact.setWordWrap(True)
+        self._vbox.addWidget(contact)
+        self._contact = QTextEdit()
+        self._contact.setTabChangesFocus(True)
+        self._contact.setAcceptRichText(False)
+        try:
+            try:
+                info = configfiles.state['general']['contact-info']
+            except KeyError:
+                self._contact.setPlaceholderText("Mail or IRC nickname")
+            else:
+                self._contact.setPlainText(info)
+        except Exception:
+            log.misc.exception("Failed to get contact information!")
+            self._contact.setPlaceholderText("Mail or IRC nickname")
+        self._vbox.addWidget(self._contact, 2)
+
     def _init_text(self):
         """Initialize the main text to be displayed on an exception.
 
         Should be extended by subclasses to set the actual text.
         """
-        self._lbl = QLabel(wordWrap=True, openExternalLinks=True,
-                           textInteractionFlags=Qt.LinksAccessibleByMouse)
+        self._lbl = QLabel()
+        self._lbl.setWordWrap(True)
+        self._lbl.setOpenExternalLinks(True)
+        self._lbl.setTextInteractionFlags(Qt.LinksAccessibleByMouse)
         self._vbox.addWidget(self._lbl)
 
     def _init_checkboxes(self):
         """Initialize the checkboxes."""
-        pass
 
     def _init_buttons(self):
         """Initialize the buttons."""
         self._btn_box = QDialogButtonBox()
         self._vbox.addWidget(self._btn_box)
 
-        self._btn_report = QPushButton("Report", default=True)
+        self._btn_report = QPushButton("Report")
+        self._btn_report.setDefault(True)
         self._btn_report.clicked.connect(self.on_report_clicked)
         self._btn_box.addButton(self._btn_report, QDialogButtonBox.AcceptRole)
 
-        self._btn_cancel = QPushButton("Don't report", autoDefault=False)
+        self._btn_cancel = QPushButton("Don't report")
+        self._btn_cancel.setAutoDefault(False)
         self._btn_cancel.clicked.connect(self.finish)
         self._btn_box.addButton(self._btn_cancel, QDialogButtonBox.RejectRole)
 
     def _init_info_text(self):
         """Add an info text encouraging the user to report crashes."""
-        info_label = QLabel("<br/><b>Note that without your help, I can't fix "
-                            "the bug you encountered.</b>",
-                            wordWrap=True)
+        info_label = QLabel("<br/>There is currently a big backlog of crash "
+                            "reports. Thus, it might take a while until your "
+                            "report is seen.<br/>A new tool allowing for more "
+                            "automation will fix this, but is not ready yet "
+                            "at this point.")
+        info_label.setWordWrap(True)
         self._vbox.addWidget(info_label)
 
     def _gather_crash_info(self):
@@ -251,12 +250,12 @@ class _CrashDialog(QDialog):
         except Exception:
             self._crash_info.append(("Launch time", traceback.format_exc()))
         try:
-            self._crash_info.append(("Version info", version.version()))
+            self._crash_info.append(("Version info", version.version_info()))
         except Exception:
             self._crash_info.append(("Version info", traceback.format_exc()))
         try:
-            conf = objreg.get('config')
-            self._crash_info.append(("Config", conf.dump_userconfig()))
+            self._crash_info.append(("Config",
+                                     config.instance.dump_userconfig()))
         except Exception:
             self._crash_info.append(("Config", traceback.format_exc()))
         try:
@@ -296,24 +295,28 @@ class _CrashDialog(QDialog):
     def _save_contact_info(self):
         """Save the contact info to disk."""
         try:
-            state = objreg.get('state-config')
-            state['general']['contact-info'] = self._contact.toPlainText()
+            info = self._contact.toPlainText()
+            configfiles.state['general']['contact-info'] = info
         except Exception:
             log.misc.exception("Failed to save contact information!")
 
-    def report(self):
-        """Paste the crash info into the pastebin."""
+    def report(self, *, info=None, contact=None):
+        """Paste the crash info into the pastebin.
+
+        If info/contact are given as arguments, they override the values
+        entered in the dialog.
+        """
         lines = []
         lines.append("========== Report ==========")
-        lines.append(self._info.toPlainText())
+        lines.append(info or self._info.toPlainText())
         lines.append("========== Contact ==========")
-        lines.append(self._contact.toPlainText())
+        lines.append(contact or self._contact.toPlainText())
         lines.append("========== Debug log ==========")
         lines.append(self._debug_log.toPlainText())
         self._paste_text = '\n\n'.join(lines)
         try:
             user = getpass.getuser()
-        except Exception as e:
+        except Exception:
             log.misc.exception("Error while getting user")
             user = 'unknown'
         try:
@@ -331,8 +334,6 @@ class _CrashDialog(QDialog):
         self._btn_report.setEnabled(False)
         self._btn_cancel.setEnabled(False)
         self._btn_report.setText("Reporting...")
-        self._paste_client.success.connect(self.on_paste_success)
-        self._paste_client.error.connect(self.show_error)
         self.report()
 
     @pyqtSlot()
@@ -368,7 +369,6 @@ class _CrashDialog(QDialog):
                          "but you're currently running v{} - please "
                          "update!".format(newest, qutebrowser.__version__))
         text = '<br/><br/>'.join(lines)
-        self.finish()
         msgbox.information(self, "Report successfully sent!", text,
                            on_finished=self.finish, plain_text=False)
 
@@ -385,7 +385,6 @@ class _CrashDialog(QDialog):
                      "<a href=https://www.qutebrowser.org/>qutebrowser.org</a> "
                      "by yourself.".format(msg))
         text = '<br/><br/>'.join(lines)
-        self.finish()
         msgbox.information(self, "Report successfully sent!", text,
                            on_finished=self.finish, plain_text=False)
 
@@ -408,8 +407,6 @@ class ExceptionCrashDialog(_CrashDialog):
     """
 
     def __init__(self, debug, pages, cmdhist, exc, qobjects, parent=None):
-        self._chk_log = None
-        self._chk_restore = None
         super().__init__(debug, parent)
         self._pages = pages
         self._cmdhist = cmdhist
@@ -429,19 +426,21 @@ class ExceptionCrashDialog(_CrashDialog):
         self._chk_restore = QCheckBox("Restore open pages")
         self._chk_restore.setChecked(True)
         self._vbox.addWidget(self._chk_restore)
-        self._chk_log = QCheckBox("Include a debug log in the report",
-                                  checked=True)
+        self._chk_log = QCheckBox("Include a debug log in the report")
+        self._chk_log.setChecked(True)
         try:
-            if config.get('general', 'private-browsing'):
+            if config.val.content.private_browsing:
                 self._chk_log.setChecked(False)
         except Exception:
             log.misc.exception("Error while checking private browsing mode")
         self._chk_log.toggled.connect(self._set_crash_info)
         self._vbox.addWidget(self._chk_log)
-        info_label = QLabel("<i>This makes it a lot easier to diagnose the "
-                            "crash, but it might contain sensitive "
-                            "information such as which pages you visited "
-                            "or keyboard input.</i>", wordWrap=True)
+        info_label = QLabel("This makes it a lot easier to diagnose the "
+                            "crash.<br/><b>Note that the log might contain "
+                            "sensitive information such as which pages you "
+                            "visited or keyboard input.</b><br/>You can show "
+                            "and edit the log above.")
+        info_label.setWordWrap(True)
         self._vbox.addWidget(info_label)
 
     def _get_error_type(self):
@@ -464,8 +463,10 @@ class ExceptionCrashDialog(_CrashDialog):
                 ("Objects", self._qobjects),
             ]
             try:
-                self._crash_info.append(
-                    ("Debug log", log.ram_handler.dump_log()))
+                text = "Log output was disabled."
+                if log.ram_handler is not None:
+                    text = log.ram_handler.dump_log()
+                self._crash_info.append(("Debug log", text))
             except Exception:
                 self._crash_info.append(("Debug log", traceback.format_exc()))
 
@@ -473,9 +474,9 @@ class ExceptionCrashDialog(_CrashDialog):
     def finish(self):
         self._save_contact_info()
         if self._chk_restore.isChecked():
-            self.accept()
+            self.done(Result.restore)
         else:
-            self.reject()
+            self.done(Result.no_restore)
 
 
 class FatalCrashDialog(_CrashDialog):
@@ -491,7 +492,6 @@ class FatalCrashDialog(_CrashDialog):
     """
 
     def __init__(self, debug, text, parent=None):
-        self._chk_history = None
         super().__init__(debug, parent)
         self._log = text
         self.setAttribute(Qt.WA_DeleteOnClose)
@@ -499,7 +499,8 @@ class FatalCrashDialog(_CrashDialog):
         self._type, self._func = parse_fatal_stacktrace(self._log)
 
     def _get_error_type(self):
-        if self._type == 'Segmentation fault':
+        if self._type in ['Segmentation fault',
+                          'Windows access violation']:
             return 'segv'
         else:
             return self._type
@@ -512,7 +513,7 @@ class FatalCrashDialog(_CrashDialog):
         text = ("<b>qutebrowser was restarted after a fatal crash.</b><br/>"
                 "<br/>Note: Crash reports for fatal crashes sometimes don't "
                 "contain the information necessary to fix an issue. Please "
-                "follow the steps in <a href='https://github.com/The-Compiler/"
+                "follow the steps in <a href='https://github.com/qutebrowser/"
                 "qutebrowser/blob/master/doc/stacktrace.asciidoc'>"
                 "stacktrace.asciidoc</a> to submit a stacktrace.<br/>")
         self._lbl.setText(text)
@@ -521,10 +522,10 @@ class FatalCrashDialog(_CrashDialog):
         """Add checkboxes to the dialog."""
         super()._init_checkboxes()
         self._chk_history = QCheckBox("Include a history of the last "
-                                      "accessed pages in the report.",
-                                      checked=True)
+                                      "accessed pages in the report.")
+        self._chk_history.setChecked(True)
         try:
-            if config.get('general', 'private-browsing'):
+            if config.val.content.private_browsing:
                 self._chk_history.setChecked(False)
         except Exception:
             log.misc.exception("Error while checking private browsing mode")
@@ -536,10 +537,31 @@ class FatalCrashDialog(_CrashDialog):
         super()._gather_crash_info()
         if self._chk_history.isChecked():
             try:
-                history = objreg.get('web-history').get_recent()
-                self._crash_info.append(("History", ''.join(history)))
+                if history.web_history is None:
+                    history_data = '<unavailable>'  # type: ignore[unreachable]
+                else:
+                    history_data = '\n'.join(str(e) for e in
+                                             history.web_history.get_recent())
             except Exception:
-                self._crash_info.append(("History", traceback.format_exc()))
+                history_data = traceback.format_exc()
+            self._crash_info.append(("History", history_data))
+
+    @pyqtSlot()
+    def on_report_clicked(self):
+        """Prevent empty reports."""
+        if (not self._info.toPlainText().strip() and
+                not self._contact.toPlainText().strip() and
+                self._get_error_type() == 'segv' and
+                self._func == 'qt_mainloop'):
+            msgbox.msgbox(parent=self, title='Empty crash info',
+                          text="Empty reports for fatal crashes are useless "
+                          "and mean I'll spend time deleting reports I could "
+                          "spend on developing qutebrowser instead.\n\nPlease "
+                          "help making qutebrowser better by providing more "
+                          "information, or don't report this.",
+                          icon=QMessageBox.Critical)
+        else:
+            super().on_report_clicked()
 
 
 class ReportDialog(_CrashDialog):
@@ -567,7 +589,6 @@ class ReportDialog(_CrashDialog):
 
     def _init_info_text(self):
         """We don't want an info text as the user wanted to report."""
-        pass
 
     def _get_error_type(self):
         return 'report'
@@ -581,7 +602,10 @@ class ReportDialog(_CrashDialog):
             ("Objects", self._qobjects),
         ]
         try:
-            self._crash_info.append(("Debug log", log.ram_handler.dump_log()))
+            text = "Log output was disabled."
+            if log.ram_handler is not None:
+                text = log.ram_handler.dump_log()
+            self._crash_info.append(("Debug log", text))
         except Exception:
             self._crash_info.append(("Debug log", traceback.format_exc()))
 
@@ -600,8 +624,10 @@ class ReportErrorDialog(QDialog):
                        "crash@qutebrowser.org</a> - Thanks!".format(
                            html.escape(exc_text)))
         vbox.addWidget(label)
-        txt = QTextEdit(readOnly=True, tabChangesFocus=True,
-                        acceptRichText=False)
+        txt = QTextEdit()
+        txt.setReadOnly(True)
+        txt.setTabChangesFocus(True)
+        txt.setAcceptRichText(False)
         txt.setText(text)
         txt.selectAll()
         vbox.addWidget(txt)
@@ -609,7 +635,7 @@ class ReportErrorDialog(QDialog):
         hbox = QHBoxLayout()
         hbox.addStretch()
         btn = QPushButton("Close")
-        btn.clicked.connect(self.close)
+        btn.clicked.connect(self.close)  # type: ignore[arg-type]
         hbox.addWidget(btn)
         vbox.addLayout(hbox)
 
@@ -630,13 +656,12 @@ def dump_exception_info(exc, pages, cmdhist, qobjects):
     print(''.join(traceback.format_exception(*exc)), file=sys.stderr)
     print("\n---- Version info ----", file=sys.stderr)
     try:
-        print(version.version(), file=sys.stderr)
+        print(version.version_info(), file=sys.stderr)
     except Exception:
         traceback.print_exc()
     print("\n---- Config ----", file=sys.stderr)
     try:
-        conf = objreg.get('config')
-        print(conf.dump_userconfig(), file=sys.stderr)
+        print(config.instance.dump_userconfig(), file=sys.stderr)
     except Exception:
         traceback.print_exc()
     print("\n---- Commandline args ----", file=sys.stderr)
